
/********************************************************************
 *
 * FILE:
 *	stChorus.c
 *
 * DESCRIPTION:
 *	Uses the API to link up to the mix engine. The API does the necessary
 *	connections to shared memory, and does a simple negotiation with the
 *	frontend to request addition to the EffectSLab frontend. The APIs are
 *	really pretty trivial, in as much as there is no flexibility the what
 *	you can do, just attach, initialise, and do read/write operations. There
 *	are no bells or whistles. There aren't even any manual pages. The up
 *	side of this approach is simplicity - the user is removed from the details
 *	of SHMEM and the specifics of the mixSLab internal memory formats. The
 *	obvious downside is the lack of flexibility. Since there are is little
 *	documentation on the APIs, here the details.
 *
 *		int * seAttach(int memID, int busID)
 *
 *			memId is the shared memory segment for the controlBuffer, and
 *			busID is the bus to which to attach. These are passed as argv[1]
 *			and argv[2] when the process is forked by the frontend.
 *			The returned value is used in subsequent calls as a connection id.
 *
 *		int seAddCtrl(int *devPtr, char *title, int type)
 *
 *			devPtr is returned by seAttach(). Title is the text inserted 
 *			under the controller for user recognition. Type, for now, is
 *			just SLab_Potmeter, but buttons, scalers and entries may be added.
 *			The integer result is used by seGetCtrlVal() later.
 *
 *		int seInit(int *devPtr, char *name, char *shortName, int (*pcheck)())
 *
 *			devPtr is from the seAttach(). Name is displayed as the effect
 *			name in the EffectsSLab GUI (truncated to 15 chars), and the
 *			shortname (7 chars), should appear on the StudioSLab, under the
 *			send pot when a track is connected to a bus. pcheck is an
 *			optional callback executed by API when paramter values change.
 *			This can be set to NULL if interface make the paramter checks
 *			itself. The callback routine is passed two parameters, the
 *			associated devPtr, and the parameter number that changed.
 *			The returned value is the fixed block size for each read and
 *			write operation.
 *
 *		int seGetOpts(int *devPtr, int option)
 *
 *			devPtr from the usual source. The option should be either
 *			SLAB_RESOLUTION or SLAB_SAMPLE_RATE. Return value is the
 *			given parameter for the current open songfile.
 *
 *		char *seGetFileName(char name[80])
 *
 *			The name of the current open songfile is returned in the name
 *			space provided.
 *
 *		int seGetCtrlVal(int *devPtr, int controllerID)
 *
 *			devPtr from the usual source, and controllerID from a
 *			previous seAddCtrl() call. Returns the current value of the
 *			given controller, as set by the user in the EffectSLab gui.
 *
 *
 *		int seSetCtrlDisp(int *devPtr, int controllerID, char *dispText)
 *
 *			devPtr from the usual source, and controllerID from a
 *			previous seAddCtrl() call. Sets a string to be displayed in the
 *			frontend for the value of this controller. This will allow the
 *			effect to set actual values and units in the display.
 *
 *		int * seRead(int *devPtr)
 *
 *			devPtr from the usual source. Returns a pointer to the next
 *			available loaded buffer from the mix engine. The buffer is in
 *			a fixed area of shared memory, and it should really be returned
 *			via an seWrite() call afterwards, although little damage will
 *			result from this not happening - the engine will log an underrun
 *			against this bus.
 *
 *		int seWrite(int *devPtr, int *busData)
 *
 *			devPtr from the usual source. The return value is not used in
 *			this release, and may never be. The busData location should be
 *			the value from the oldest outstanding read operation, but it is
 *			not heavily checked. There may be a minimum of 2 outstanding
 *			seRead() request, with the maximum being platform dependent -
 *			2 on SunOS, 4 on Linux.
 *
 *		int seDetach(int *devPtr)
 *
 *			Detaches itself from shared memory, and relinquishes records in 
 *			the controlBuffer internal data structures. Returns 0 on success
 *			and a negative integer on failure.
 *
 * TOC:
 *	main()
 *	chorus()
 *	checkParams()
 *
 *******************************************************************/

#include "slabEffects.h"

/*
 * For historic reasons, the descriptor for bus attachments is int *.
 */
int *devPtr;
int *devPtr2;

/*
 * Used by the user interface in the effects rack. Note that the EFFECT_NAME
 * should refer to the binary executable file name, since it is used by the
 * GUI for baselining. This name will be used to restart the effect during
 * reloading of this file into the studio.
 */
#define EFFECT_NAME "stChorus"
#define SHORT_NAME "Chorus"

/*
 * These are dependant on the effect - parameters etc. This chorus implements
 * 4 individual delay lines, with a MAX delay, at 16384 samples, or about
 * 400ms at 44100sps.
 */
#define HISTORY_SIZE 65536
#define CHORUS_SIZE 16384
#define INDEX1_START 0
#define INDEX2_START 16384
#define INDEX3_START 32768
#define INDEX4_START 49152

int history[HISTORY_SIZE];
int indexOut1 = 0;
int indexOut2 = 0;
int indexOut3 = 0;
int indexOut4 = 0;
int indexIn = 0;
int Speed1, Speed2, Speed3, Speed4;

int speed, breadth;
int Breadth = 128;

int currentBreadth1 = 0;
int currentBreadth2 = 0;
int currentBreadth3 = 0;
int currentBreadth4 = 0;
int countBreadth1 = 0;
int countBreadth2 = 0;
int countBreadth3 = 0;
int countBreadth4 = 0;
int changeBreadth1 = 1;
int changeBreadth2 = 1;
int changeBreadth3 = 1;
int changeBreadth4 = 1;

int Min = 1;

int checkParams(), sampleRate;

/*
 * When the user interface attaches an effect, it passes two parameters, no
 * switches. This is very rudementary, but will be extended:
 *
 * argv[1] is the controlbuffer segID, argv[2] should be the bus requested
 */
main(argc, argv)
int argc;
char *argv[];
{
	int *busData, *busData2, count, i;

	if (argc < 3)
	{
		printf("%s: parameter problem\n", argv[0]);
		exit(-1);
	}

	/*
	 * seAttach() will register with the mixengine, and link up the shared
	 * memory control and data structures. It effectively reserves a bus.
	 * The SLAB_EFFECTS_MONO really refers to the first bus attachment. We
	 * can attach to a second one with the SLAB_EFFECTS_STEREO flag. It is
	 * only used as a switch to use a slightly different link up protocol.
	 * If we use the STEREO bus link, we will be given access to the configured
	 * bus plus 1. This is off little interest to the effect, but is needed
	 * to remix the return data back into the main mix.
	 */
	if ((devPtr = seAttach(atoi(argv[1]), atoi(argv[2]),
		SLAB_EFFECTS_MONO)) == (int *) NULL)
	{
		printf("%s: could not attach to requested bus\n", argv[0]);
		exit(-2);
	}

	if ((devPtr2 = seAttach(atoi(argv[1]), atoi(argv[2]),
		SLAB_EFFECTS_STEREO)) == (int *) NULL)
	{
		printf("%s: could not attach to second bus\n", argv[0] + 1);
		seDetach(devPtr);
		exit(-2);
	}

	/*
	 * At this point we can use seAddCtrl() to request up to 8 parameters
	 * which will be shown in the SLab effects user interface. The chorus algo
	 * requires speed, breadth. The calls return an int which is
	 * used later when requesting paramter values.
	 *
	 * Interface only supports the SLab_Potmeter controller type now.
	 */
	speed = seAddCtrl(devPtr, "Speed", SLab_Potmeter);
	breadth = seAddCtrl(devPtr, "Breadth", SLab_Potmeter);

	/*
	 * seInit() returns our fixed buffer size. We always use a fixed buffer as
	 * an internal data representation. No exceptions. Data is manipulated in
	 * this fixed buffer, and then the buffer (pointer) is returned to the bus.
	 *
	 * The buffer size is configurable in the GUI, to allow some memory tuning.
	 * For this reason we need to be able to find out the current value.
	 */
	count = seInit(devPtr, EFFECT_NAME, SHORT_NAME, checkParams);

	/*
	 * Init the display
	 */
	sampleRate = seGetOpts(devPtr, SLAB_SAMPLE_RATE);
	checkParams(devPtr, 0);

	/*
	 * This is a forever loop, sort of.....
	 * THe Chorus algorithm only uses the two busses to return data. It does
	 * accept data from the second bus - but it is a mono to stereo algorithm,
	 * so the second bus is just overwritten.
	 */
	while ((busData = seRead(devPtr)) != (int *) -1) {
		busData2 = seRead(devPtr2);
		chorus(busData, busData2, count);
		seWrite(devPtr, busData);
		seWrite(devPtr2, busData2);
	}

	/*
	 * seDetach() will deregister with the mixengine, and remove links to the
	 * shared memory control and data structures.
	 */
	if (seDetach(devPtr) < 0)
		printf("%s: could not detach from requested bus\n", argv[0]);
	if (seDetach(devPtr2) < 0)
	{
		printf("%s: could not detach from second bus\n", argv[0] + 1);
		exit(-3);
	}

	exit(0);
}

/*
 * This is the actual chorus algorithm. The rest of this file, excluding the
 * variable definitions, is basically an effects template. If we want to have
 * a different stereo effect, the basics of what we need to change is below:
 */
chorus(busData, busData2, count)
register int *busData, *busData2, count;
{
	register int holder;

	for (; count > 0; count--)
	{
		history[indexIn] = 
		history[indexIn + INDEX2_START] =
		history[indexIn + INDEX3_START] =
		history[indexIn + INDEX4_START] = *busData;

		/*
		 * Fill the second output bus - overwrite contents
		 */
		*(busData2++) = history[indexOut1] +
			history[indexOut3 + INDEX3_START];

		/*
		 * Similarly the first bus
		 */
		*(busData++) = history[indexOut2] +
			history[indexOut4 + INDEX4_START];

		if (++countBreadth1 >= Speed1)
		{
			countBreadth1 = 0;

			if ((currentBreadth1 += changeBreadth1) >= Breadth)
				changeBreadth1 = -1;
			if (currentBreadth1 <= Min)
				changeBreadth1 = 1;

			/*
			 * See if we are closing our window, skip a sample.
			 */
			if (changeBreadth1 < 0) {
				if ((indexOut1 += 1) >= Breadth) indexOut1 = 0;
				if ((indexOut1 += 1) >= Breadth) indexOut1 = 0;
			}
		} else {
			if (++indexOut1 >= Breadth) indexOut1 = 0;
		}

		if (++countBreadth2 >= Speed2)
		{
			countBreadth2 = 0;

			if ((currentBreadth2 += changeBreadth2) >= Breadth)
				changeBreadth2 = -1;
			if (currentBreadth2 <= Min)
				changeBreadth2 = 1;
			/*
			 * See if we are closing our window, skip a sample.
			 */
			if (changeBreadth2 < 0) {
				if ((indexOut2 += 1) >= Breadth) indexOut2 = 0;
				if ((indexOut2 += 1) >= Breadth) indexOut2 = 0;
			}
		} else {
			if (++indexOut2 >= Breadth) indexOut2 = 0;
		}

		if (++countBreadth3 >= Speed3)
		{
			countBreadth3 = 0;

			if ((currentBreadth3 += changeBreadth3) >= Breadth)
				changeBreadth3 = -1;
			if (currentBreadth3 <= Min)
				changeBreadth3 = 1;

			/*
			 * See if we are closing our window, skip a sample.
			 */
			if (changeBreadth3 < 0) {
				if ((indexOut3 += 1) >= Breadth) indexOut3 = 0;
				if ((indexOut3 += 1) >= Breadth) indexOut3 = 0;
			}
		} else {
			if (++indexOut3 >= Breadth) indexOut3 = 0;
		}

		if (++countBreadth4 >= Speed4)
		{
			countBreadth4 = 0;

			if ((currentBreadth4 += changeBreadth4) >= Breadth)
				changeBreadth4 = -1;
			if (currentBreadth4 <= Min)
				changeBreadth4 = 1;

			/*
			 * See if we are closing our window, skip a sample.
			 */
			if (changeBreadth4 < 0) {
				if ((indexOut4 += 1) >= Breadth) indexOut4 = 0;
				if ((indexOut4 += 1) >= Breadth) indexOut4 = 0;
			}
		} else {
			if (++indexOut4 >= Breadth) indexOut4 = 0;
		}

		if (++indexIn >= Breadth) indexIn = 0;
	} 
}

checkParams(devPtr, param)
int *devPtr, param;
{
	char text[32];

	if (param == speed) {
		Speed1 = (256 - seGetCtrlVal(devPtr, speed)) << 2;
		Speed2 = Speed1 + 10;
		Speed3 = Speed1 + 20;
		Speed4 = Speed1 + 30;

		/*
		 * The actual speed here is a function of both Speed and Breadth.
		 */
		sprintf(text, "Speed: %2.2f Hz",
			(float) sampleRate / (Speed1 * Breadth));
		seSetCtrlDisp(devPtr, speed, text);
		return;
	}

	Breadth = (seGetCtrlVal(devPtr, breadth) << 4) + 1;

	sprintf(text, "Breadth: %2.2fms", (float) Breadth * 1000 / sampleRate);
	seSetCtrlDisp(devPtr, breadth, text);
}

